
#include <string.h>

#include "common.h"
#include "dump.h"
#include "xml.h"
#include "png.h"
#include "objects.h"

char indent[11] = "\t\t\t\t\t\t\t\t\t\t";
#define INDENT (indent+10-indentation)

uint32_t getvalue(uint8_t *data, int size)
{
  uint32_t value = *(uint32_t*)(data);
  value &= 0xffffffff >> (32-8*size);

  return value;
}

int32_t getsvalue(uint8_t *data, int size)
{
  int32_t value = *(int32_t*)(data);
  value = (value << (32-8*size)) >> (32-8*size);

  return value;
}


uint32_t dumpflags(FILE *xml, uint8_t *data, const char **flags, int size, int indentation)
{
  uint32_t value, i, state;
  const char *name;
  char defname[32];

  if (size > 4) die("Can't dump flags with %d bytes", size);

  value = getvalue(data, size);

  for (i=0; i<size*8; i++) {
	if (*flags) {
		name = *flags;
		flags++;
	} else
		name = "";

	if (!strlen(name)) {
		sprintf(defname, "bit%d", i);
		name = defname;
	}
	state = (value & (1 << i)) >> i;
	if ((name != defname) || state)
		fprintf(xml, "%s<bit name=\"%s\">%d</bit>\n", INDENT, name, state);
  }
  return size;
}

uint32_t dodump(FILE *xml, uint8_t *data, varinf *vars, int indentation)
{
  int i, j;
  uint32_t dumped, totaldumped, ofs;
  const char *name, *fname;
  char defname[32], arrname[32];

  totaldumped = 0;
  ofs = 0;

  for (i=0; vars[i].ofs != -1; i++) {
	fname = vars[i].name;
	if (!strlen(fname)) {
		sprintf(defname, "field_%X", vars[i].ofs);
		fname=defname;
	}
	if (ofs != vars[i].ofs) die("Structure is invalid at %X", ofs);
	for (j=0; j<vars[i].num; j++) {
		name = fname;
		if (vars[i].num > 1) {
			sprintf(arrname, "%s[%d]", fname, j);
			name = arrname;
		}

		if (vars[i].flags) {
			fprintf(xml, "%s<bitmask name=\"%s\">\n", INDENT, name);
			dumped = dumpflags(xml, data+ofs, vars[i].flags, vars[i].size, indentation+1);
			fprintf(xml, "%s</bitmask>\n", INDENT);
		} else if (vars[i].structvars) {
			fprintf(xml, "%s<structure name=\"%s\">\n", INDENT, name);
			dumped = dodump(xml, data+ofs, vars[i].structvars,
				indentation+1);

			fprintf(xml, "%s</structure>\n", INDENT);
		} else if (strlen(vars[i].name)) {
			fprintf(xml, "%s<variable name=\"%s\" size=\"%X\">%d</variable>\n", INDENT, name, vars[i].size, getvalue(data+ofs, vars[i].size));
			dumped = vars[i].size;
		} else {
			uint32_t value = getvalue(data+ofs, vars[i].size);
			if (value)
				fprintf(xml, "%s<unknown name=\"%s\" size=\"%X\">%d</unknown>\n", INDENT, name, vars[i].size, value);
			dumped = vars[i].size;
		}
		ofs += dumped;
		totaldumped += dumped;
	}
  }
  return totaldumped;
}

uint32_t dumpvars(FILE *xml, uint8_t *data, varinf *vars)
{
  return dodump(xml, data, vars, 1);
}

uint32_t dumplang(FILE *xml, uint8_t *data)
{
  int language;
  uint32_t ofs;

  ofs = 0;
  while ( (language = data[ofs++]) != 0xff ) {
	fprintf(xml, "\t<description language=\"%d\">", language);
	putxmlstr(xml, data+ofs, 0);
	fputs("</description>\n", xml);
	ofs += strlen(data+ofs) + 1;
  }

  return ofs;
}

uint32_t dumpuseobj(FILE *xml, uint8_t *data, int num, int total, const char *type, int class)
{
  char arrtype[32];

  if (total > 1) {
	sprintf(arrtype, "%s[%d]", type, num);
	type = arrtype;
  }

  if (class != data[0]) die("Not correct class for %s; want %d got %d", type, class, data[0]);

  fprintf(xml, "\t<useobject desc=\"%s\" class=\"%d\">%.8s</useobject>\n",
	type, class, data+4);

  return 16;
}

uint32_t dumpcap(FILE *xml, uint8_t *data, int num, int total)
{
  uint32_t dumped;
  int capacity, cargotype, refcap;

  if (total);

  capacity = data[0];

  dumped = 1;

  if (capacity) {
	fprintf(xml, "\t<cargo num=\"%d\" capacity=\"%d\">\n", num, capacity);
	while (getvalue(data+dumped, 2) != 0xffff) {
		cargotype = data[dumped++];

		refcap = getvalue(data+dumped, 2);
		dumped += 2;

		fprintf(xml, "\t\t<cargotype id=\"%d\">%d</cargotype>\n", cargotype, refcap);
	}
	dumped += 2;
	fputs("\t</cargo>\n", xml);
  }

  return dumped;
}

uint32_t dumpsprites(FILE *xml, uint8_t *data)
{
  uint32_t i, j, num, size, totalsize, dumped;
  uint8_t *spritedata;
  FILE *png = NULL;

  num = getvalue(data, 4);
  size = getvalue(data+4, 4);
  dumped = 8;

  spritedata = data+8+16*num;
  totalsize = 0;

  for (i=0; i<num; i++) {
	uint32_t ofs,width,height,flags, spritesize;
	int32_t xofs,yofs;

	ofs=getvalue(data+dumped,4);  dumped+=4;
	width=getvalue(data+dumped,2);  dumped+=2;
	height=getvalue(data+dumped,2);  dumped+=2;
	xofs=getsvalue(data+dumped,2);  dumped+=2;
	yofs=getsvalue(data+dumped,2);  dumped+=2;
	flags=getvalue(data+dumped,4);  dumped+=4;

	fprintf(xml, "\t<sprite id=\"%d\" xofs=\"%d\" yofs=\"%d\">\n", i, xofs, yofs);
	dumpflags(xml, (uint8_t*) &flags, spriteflags, 4, 2);

	if (embedpng) {
		fputs("\t\t<pngdata>", xml);
	} else if (allinone) {
		// fprintf(xml, "\t\t<png x= y= width= height=>%s</png>\n", );
	} else {
		char *pngname;

		pngname = my_alloc(strlen(filebase)+16, "pngname");
		sprintf(pngname, "%d", num);
		j = strlen(pngname);
		strcpy(pngname, filebase);
		sprintf(pngname+strlen(pngname), "_%0*d.png", j, i);
		png = fopen(pngname, "wb");
		if (!png) die("Can't write %s: %s", pngname, strerror(errno));

		fprintf(xml, "\t\t<pngfile>%s</pngfile>\n", pngname);

		free(pngname);
	}

	spritesize = makepng(xml, png, spritedata + ofs, width, height, flags);

	if (embedpng) {
		fputs("</pngdata>\n", xml);
	} else if (!allinone) {
		fclose(png);
	}

	totalsize += spritesize;

	fputs("\t</sprite>\n", xml);
  }

  dumped += totalsize;
  if (totalsize != size) die("Sprite data different from size; wanted %d had %d", size, totalsize);

  return dumped;
}

int dumpchunk(FILE *xml, uint8_t *data, int class)
{
  uint32_t dumped;

  if ( (class < sizeof(dumpprocs)) && dumpprocs[class])
	return dumpprocs[class](xml, data);
  else
	die("Can't translate class %02X yet", class);

  return -1;
}

