#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <png.h>

#include "common.h"
#include "dump.h"

uint32_t putspriterow(uint8_t *data, int flags, int y, int width, png_byte *row)
{
  uint32_t offset, last, ofs, len;

  memset(row, 0, width);

  if (flags & 4) {
	offset=getvalue(data+2*y, 2);

	do {
		last=data[offset]   & 0x80;
		len=data[offset++] & 0x7f;
		ofs=data[offset++];

		memcpy(row+ofs, data+offset, len);

		offset+=len;
	} while (!last);
  } else {
	memcpy(row, data+y*width, width);
	offset = (y+1)*width;
  }
  return offset;
}

void xml_write(png_structp png_ptr, png_bytep data, png_size_t size)
{
//  putxmlstr(png_ptr->io_ptr, data, iso_to_utf8, size);
  fwrite(data, 1, size, png_ptr->io_ptr);
}

void xml_flush(png_structp png_ptr)
{
  fflush(png_ptr->io_ptr);
}

#include "lomopal.h"

uint32_t makepng(FILE *xml, FILE *png, uint8_t *data, uint32_t width, uint32_t height, uint32_t flags)
{
   png_structp png_ptr;
   png_infop info_ptr;
   png_colorp palette;
   png_byte *row;

   int i,j;
   uint32_t dumped;

   dumped = 0;

#if !defined(__BORLANDC__) || __BORLANDC__ > 0x500

//   fputs("<![CDATA[", xml);

   if (setmode(fileno(xml), O_BINARY) == 1) die("Can't switch to binary file mode");

	// create the png structures
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   if (!png_ptr) die("Error creating png write struct");

   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
	png_destroy_write_struct(&png_ptr, png_infopp_NULL);
	die("Error creating png info struct");
   }

	// error handlingg
   if (setjmp(png_jmpbuf(png_ptr))) {
		// if we get here, we had a problem writing the file
	png_destroy_write_struct(&png_ptr, &info_ptr);
	die("Error while writing png data");
   }

   if (embedpng) {
		// use our own write functions
	png_set_write_fn(png_ptr, xml, xml_write, xml_flush);
   } else if (allinone) {
   } else {
	png_init_io(png_ptr, png);
   }

	// set image dimensions
   png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_PALETTE,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// set the palette
   palette = (png_colorp)png_malloc(png_ptr, PNG_MAX_PALETTE_LENGTH * sizeof (png_color));
   if (!palette) die("Not enough memory for %s, wanted %d", "palette", PNG_MAX_PALETTE_LENGTH * sizeof (png_color));

   for (i=0; i<256; i++) {
	palette[i].red = lomopalette[i][0];
	palette[i].green = lomopalette[i][1];
	palette[i].blue = lomopalette[i][2];
   }

   png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);

	// set the gamma
   png_set_gAMA(png_ptr, info_ptr, 1.4);

	// write the png header
   png_write_info(png_ptr, info_ptr);

   if (embedpng || !allinone) {

   row = my_alloc(width, "row");

	// write the image line by line
   for (i=0; i<height; i++) {
	uint32_t offset = putspriterow(data, flags, i, width, row);

	if (offset > dumped) dumped=offset;

	png_write_row(png_ptr, row);
   }

   free(row);

   }

	// done writing, clean up
   png_write_end(png_ptr, info_ptr);
   png_free(png_ptr, palette);
   png_destroy_write_struct(&png_ptr, &info_ptr);

   if (setmode(fileno(xml), O_TEXT) == 1) die("Can't switch to text file mode");
//   fputs("]]>", xml);

#endif
   return dumped;
}

uint32_t readpng(const char *name, uint8_t **data, uint32_t *width, uint32_t *height)
{
  return 0;
}
